# sage_setup: distribution = sagemath-glpk
#*****************************************************************************
#       Copyright (C) 2010 Nathann Cohen <nathann.cohen@gmail.com>
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

cdef extern from "glpk.h":
    # constants for smcp control
    int GLP_MSG_OFF
    int GLP_MSG_ERR
    int GLP_MSG_ON
    int GLP_MSG_ALL

    int GLP_PRIMAL
    int GLP_DUALP
    int GLP_DUAL

    int GLP_PT_STD
    int GLP_PT_PSE

    int GLP_RT_STD
    int GLP_RT_HAR

    int GLP_ON
    int GLP_OFF

    # constants for iocp control, not already in simplex
    int GLP_BR_FFV
    int GLP_BR_LFV
    int GLP_BR_MFV
    int GLP_BR_DTH
    int GLP_BR_PCH

    int GLP_BT_DFS
    int GLP_BT_BFS
    int GLP_BT_BLB
    int GLP_BT_BPH

    int GLP_PP_NONE
    int GLP_PP_ROOT
    int GLP_PP_ALL

    # error codes
    int GLP_EBADB
    int GLP_ESING
    int GLP_ECOND
    int GLP_EBOUND
    int GLP_EFAIL
    int GLP_EOBJLL
    int GLP_EOBJUL
    int GLP_EITLIM
    int GLP_ETMLIM
    int GLP_ENOPFS
    int GLP_ENODFS
    int GLP_EROOT
    int GLP_ESTOP
    int GLP_EMIPGAP
    int GLP_EDATA
    int GLP_ERANGE

    int GLP_UNDEF
    int GLP_OPT
    int GLP_FEAS
    int GLP_NOFEAS
    int GLP_INFEAS
    int GLP_UNBND

    # other constants
    int GLP_MAX
    int GLP_MIN
    int GLP_UP
    int GLP_FR
    int GLP_DB
    int GLP_FX
    int GLP_LO
    int GLP_CV
    int GLP_IV
    int GLP_BV
    int GLP_MPS_DECK
    int GLP_MPS_FILE

    int GLP_MSG_DBG

    int GLP_BS      # basic variable
    int GLP_NL      # non-basic variable on lower bound
    int GLP_NU      # non-basic variable on upper bound
    int GLP_NF      # non-basic free (unbounded) variable
    int GLP_NS      # non-basic fixed variable



# sage_setup: distribution = sagemath-lcalc
# distutils: libraries = M_LIBRARIES NTL_LIBRARIES Lfunction
# distutils: extra_compile_args = NTL_CFLAGS -O3 -ffast-math
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++
r"""
Rubinstein's lcalc library

This is a wrapper around Michael Rubinstein's lcalc.
See http://oto.math.uwaterloo.ca/~mrubinst/L_function_public/CODE/.

AUTHORS:

- Rishikesh (2010): added compute_rank() and hardy_z_function()
- Yann Laigle-Chapuy (2009): refactored
- Rishikesh (2009): initial version
"""

# ****************************************************************************
#       Copyright (C) 2009 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.signals cimport sig_on, sig_off

from sage.cpython.string cimport str_to_bytes

from sage.libs.gmp.mpz cimport *
from sage.libs.mpfr cimport *
from sage.rings.integer cimport Integer

from sage.rings.complex_mpfr cimport ComplexNumber
from sage.rings.complex_mpfr import ComplexField
CCC = ComplexField()

from sage.rings.real_mpfr cimport RealNumber
from sage.rings.real_mpfr import RealField
RRR = RealField()
pi = RRR.pi()

initialize_globals()

##############################################################################
# Lfunction: base class for L-functions
##############################################################################

cdef class Lfunction:
    # virtual class
    def __init__(self, name, what_type_L, dirichlet_coefficient,
                 period, Q, OMEGA, gamma, lambd, pole, residue):
        """
        Initialization of `L`-function objects.
        See derived class for details, this class is not supposed to be
        instantiated directly.

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: Lfunction_from_character(DirichletGroup(5)[1])
            L-function with complex Dirichlet coefficients
        """
        cdef int i              # for indexing loops
        cdef Integer tmpi       # for accessing integer values
        cdef RealNumber tmpr    # for accessing real values
        cdef ComplexNumber tmpc # for accessing complex values

        _name = str_to_bytes(name)
        cdef char *NAME = _name
        cdef int what_type = what_type_L

        tmpi = Integer(period)
        cdef int Period = mpz_get_si(tmpi.value)
        tmpr = RRR(Q)
        cdef double q=mpfr_get_d(tmpr.value, MPFR_RNDN)
        tmpc = CCC(OMEGA)
        cdef c_Complex w=new_Complex(mpfr_get_d(tmpc.__re, MPFR_RNDN), mpfr_get_d(tmpc.__im, MPFR_RNDN))

        cdef int A=len(gamma)
        cdef double *g=new_doubles(A+1)
        cdef c_Complex *l=new_Complexes(A+1)
        for i from 0 <= i < A:
            tmpr = RRR(gamma[i])
            g[i+1] = mpfr_get_d(tmpr.value, MPFR_RNDN)
            tmpc = CCC(lambd[i])
            l[i+1] = new_Complex(mpfr_get_d(tmpc.__re, MPFR_RNDN), mpfr_get_d(tmpc.__im, MPFR_RNDN))

        cdef int n_poles = len(pole)
        cdef c_Complex *p = new_Complexes(n_poles +1)
        cdef c_Complex *r = new_Complexes(n_poles +1)
        for i from 0 <= i < n_poles:
            tmpc=CCC(pole[i])
            p[i+1] = new_Complex(mpfr_get_d(tmpc.__re, MPFR_RNDN), mpfr_get_d(tmpc.__im, MPFR_RNDN))
            tmpc=CCC(residue[i])
            r[i+1] = new_Complex(mpfr_get_d(tmpc.__re, MPFR_RNDN), mpfr_get_d(tmpc.__im, MPFR_RNDN))

        self._init_fun(NAME, what_type, dirichlet_coefficient, Period, q,  w,  A, g, l, n_poles, p, r)

        if name:
            name += ': '

        self._repr = name + 'L-function'

        del_doubles(g)
        del_Complexes(l)
        del_Complexes(p)
        del_Complexes(r)

    def __repr__(self):
        """
        Return string representation of this `L`-function.

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: Lfunction_from_character(DirichletGroup(5)[1])
            L-function with complex Dirichlet coefficients

            sage: Lfunction_Zeta()
            The Riemann zeta function
        """
        return self._repr

    def value(self, s, derivative=0):
        """
        Compute the value of the `L`-function at ``s``.

        INPUT:

        - ``s`` -- a complex number
        - ``derivative`` -- integer (default: 0);  the derivative to be evaluated
        - ``rotate`` -- boolean (default: ``False``); if True, this returns the value of the
          Hardy Z-function (sometimes called the Riemann-Siegel Z-function or
          the Siegel Z-function)

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[2] # This is a quadratic character
            sage: L = Lfunction_from_character(chi, type='int')
            sage: (L.value(0.5) - 0.231750947504016).abs() < 1e-8
            True
            sage: v = L.value(0.2 + 0.4*I)
            sage: (v - (0.102558603193 + 0.190840777924*I)).abs() < 1e-8
            True
            sage: L = Lfunction_from_character(chi, type='double')
            sage: (L.value(0.6) - 0.274633355856345).abs() < 1e-8
            True
            sage: v = L.value(0.6 + I)
            sage: (v - (0.362258705721 + 0.43388825062*I)).abs() < 1e-8
            True

        ::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[1]
            sage: L = Lfunction_from_character(chi, type='complex')
            sage: v = L.value(0.5)
            sage: (v - (0.763747880117 + 0.21696476751*I)).abs() < 1e-8
            True
            sage: v = L.value(0.6 + 5*I)
            sage: (v - (0.702723260619 - 1.10178575243*I)).abs() < 1e-8
            True

        ::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: L = Lfunction_Zeta()
            sage: (L.value(0.5) + 1.46035450880).abs() < 1e-8
            True
            sage: v = L.value(0.4 + 0.5*I)
            sage: (v - (-0.450728958517 - 0.780511403019*I)).abs() < 1e-8
            True
        """
        cdef ComplexNumber complexified_s = CCC(s)
        cdef c_Complex z = new_Complex(mpfr_get_d(complexified_s.__re, MPFR_RNDN), mpfr_get_d(complexified_s.__im, MPFR_RNDN))
        cdef c_Complex result = self._value(z, derivative)
        return CCC(result.real(),result.imag())

    def hardy_z_function(self, s):
        """
        Compute the Hardy Z-function of the `L`-function at s.

        INPUT:

        - ``s`` -- a complex number with imaginary part between -0.5 and 0.5

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[2]  # Quadratic character
            sage: L = Lfunction_from_character(chi, type='int')
            sage: (L.hardy_z_function(0) - 0.231750947504).abs() < 1e-8
            True
            sage: L.hardy_z_function(0.5).imag().abs() < 1e-8
            True

        ::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[1]
            sage: L = Lfunction_from_character(chi, type='complex')
            sage: (L.hardy_z_function(0) - 0.793967590477).abs() < 1e-8
            True
            sage: L.hardy_z_function(0.5).imag().abs() < 1e-8
            True

        ::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: E = EllipticCurve([-82,0])
            sage: L = Lfunction_from_elliptic_curve(E, number_of_coeffs=40000)
            sage: (L.hardy_z_function(2.1) - (-0.006431791768)).abs() < 1e-8
            True
        """
        #This takes s -> .5 + I*s
        cdef ComplexNumber complexified_s = CCC(0.5)+ CCC(0,1)*CCC(s)
        cdef c_Complex z = new_Complex(mpfr_get_d(complexified_s.__re, MPFR_RNDN), mpfr_get_d(complexified_s.__im, MPFR_RNDN))
        cdef c_Complex result = self._hardy_z_function(z)
        return CCC(result.real(),result.imag())

    def compute_rank(self):
        """
        Compute the analytic rank (the order of vanishing at the center) of
        of the `L`-function.

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[2] # This is a quadratic character
            sage: L = Lfunction_from_character(chi, type='int')
            sage: L.compute_rank()
            0

        ::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: E = EllipticCurve([-82,0])
            sage: L = Lfunction_from_elliptic_curve(E, number_of_coeffs=40000)
            sage: L.compute_rank()
            3
        """
        return self._compute_rank()

    def _N(self, T):
        """
        Compute the number of zeroes upto height `T` using the formula for
        `N(T)` with the error of `S(T)`. Please do not use this. It is only
        for debugging

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[2] #This is a quadratic character
            sage: L=Lfunction_from_character(chi, type='complex')
            sage: L._N(10) # abs tol 1e-8
            4.0
        """
        cdef RealNumber real_T=RRR(T)
        cdef double double_T = mpfr_get_d(real_T.value, MPFR_RNDN)
        cdef double res_d = self._typedN(double_T)
        return RRR(res_d)

    def find_zeros(self, T1, T2, stepsize):
        """
        Finds zeros on critical line between ``T1`` and ``T2`` using step size
        of stepsize. This function might miss zeros if step size is too
        large. This function computes the zeros of the `L`-function by using
        change in signs of areal valued function whose zeros coincide with
        the zeros of `L`-function.

        Use :meth:`find_zeros_via_N` for slower but more rigorous computation.

        INPUT:

        - ``T1`` -- a real number giving the lower bound
        - ``T2`` -- a real number giving the upper bound
        - ``stepsize`` -- step size to be used for the zero search

        OUTPUT: list of the imaginary parts of the zeros which were found

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[2] # This is a quadratic character
            sage: L = Lfunction_from_character(chi, type='int')
            sage: L.find_zeros(5,15,.1)
            [6.64845334472..., 9.83144443288..., 11.9588456260...]
            sage: L = Lfunction_from_character(chi, type='double')
            sage: L.find_zeros(1,15,.1)
            [6.64845334472..., 9.83144443288..., 11.9588456260...]

        ::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[1]
            sage: L = Lfunction_from_character(chi, type='complex')
            sage: L.find_zeros(-8,8,.1)
            [-4.13290370521..., 6.18357819545...]

        ::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: L = Lfunction_Zeta()
            sage: L.find_zeros(10,29.1,.1)
            [14.1347251417..., 21.0220396387..., 25.0108575801...]
        """
        cdef doublevec result
        cdef int i
        cdef RealNumber real_T1 = RRR(T1)
        cdef RealNumber real_T2 = RRR(T2)
        cdef RealNumber real_stepsize = RRR(stepsize)
        sig_on()
        self._find_zeros_v(mpfr_get_d(real_T1.value, MPFR_RNDN),
                           mpfr_get_d(real_T2.value, MPFR_RNDN),
                           mpfr_get_d(real_stepsize.value, MPFR_RNDN), &result)
        sig_off()
        i = result.size()
        returnvalue = []
        for i in range(result.size()):
            returnvalue.append(RRR(result.ind(i)))
        result.clear()
        return returnvalue

    #The default values are from L.h. See L.h
    def find_zeros_via_N(self, count=0, start=0, max_refine=1025, rank=-1):
        """
        Find ``count`` zeros (in order of increasing magnitude) and output
        their imaginary parts. This function verifies that no zeros
        are missed, and that all values output are indeed zeros.

        If this `L`-function is self-dual (if its Dirichlet coefficients
        are real, up to a tolerance of 1e-6), then only the zeros with
        positive imaginary parts are output. Their conjugates, which
        are also zeros, are not output.

        INPUT:

        - ``count`` -- number of zeros to be found
        - ``start`` -- (default: 0) how many initial zeros to skip
        - ``max_refine`` -- when some zeros are found to be missing, the step
          size used to find zeros is refined. max_refine gives an upper limit
          on when lcalc should give up. Use default value unless you know
          what you are doing.
        - ``rank`` -- integer (default: -1); analytic rank of the `L`-function.
          If -1 is passed, then we attempt to compute it. (Use default if in
          doubt)

        OUTPUT: list of the imaginary parts of the zeros that have been found

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[2] #This is a quadratic character
            sage: L = Lfunction_from_character(chi, type='int')
            sage: L.find_zeros_via_N(3)
            [6.64845334472..., 9.83144443288..., 11.9588456260...]
            sage: L = Lfunction_from_character(chi, type='double')
            sage: L.find_zeros_via_N(3)
            [6.64845334472..., 9.83144443288..., 11.9588456260...]

        ::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[1]
            sage: L = Lfunction_from_character(chi, type='complex')
            sage: zeros = L.find_zeros_via_N(3)
            sage: (zeros[0] - (-4.13290370521286)).abs() < 1e-8
            True
            sage: (zeros[1]  - 6.18357819545086).abs() < 1e-8
            True
            sage: (zeros[2]  - 8.45722917442320).abs() < 1e-8
            True

        ::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: L = Lfunction_Zeta()
            sage: L.find_zeros_via_N(3)
            [14.1347251417..., 21.0220396387..., 25.0108575801...]
        """

        # This is the default value for message_stamp, but we have to
        # pass it explicitly since we're passing in the next argument,
        # our &result pointer.
        cdef const char* message_stamp = ""
        cdef doublevec result
        sig_on()
        self._find_zeros(count, start, max_refine, rank, message_stamp, &result)
        sig_off()
        returnvalue = []
        for i in range(result.size()):
            returnvalue.append(  RRR(result.ind(i)))
        result.clear()
        return returnvalue

    # Needs to be overridden
    cdef void _init_fun(self, char *NAME, int what_type, dirichlet_coeff, long long Period, double q,  c_Complex w, int A, double *g, c_Complex *l, int n_poles, c_Complex *p, c_Complex *r) noexcept:
        raise NotImplementedError

    cdef c_Complex _value(self, c_Complex s, int derivative) noexcept:
        raise NotImplementedError

    cdef c_Complex _hardy_z_function(self, c_Complex s) noexcept:
        raise NotImplementedError

    cdef int _compute_rank(self) noexcept:
        raise NotImplementedError

    cdef double _typedN(self, double T) noexcept:
        raise NotImplementedError

    cdef void _find_zeros_v(self, double T1, double T2, double stepsize, doublevec *result) noexcept:
        raise NotImplementedError

    cdef int _find_zeros(self, long count, long start, double max_refine, int rank, const char* message_stamp, doublevec *result) noexcept:
        raise NotImplementedError

##############################################################################
# Lfunction_I: L-functions with integer Dirichlet Coefficients
##############################################################################

cdef class Lfunction_I(Lfunction):
    r"""
    The ``Lfunction_I`` class is used to represent `L`-functions
    with integer Dirichlet Coefficients. We assume that `L`-functions
    satisfy the following functional equation.

    .. MATH::

        \Lambda(s) = \omega Q^s \overline{\Lambda(1-\bar s)}

    where

    .. MATH::

        \Lambda(s) = Q^s \left( \prod_{j=1}^a \Gamma(\kappa_j s + \gamma_j) \right) L(s)

    See (23) in :arxiv:`math/0412181`

    INPUT:

    - ``what_type_L`` -- integer; this should be set to 1 if the coefficients
      are periodic and 0 otherwise

    - ``dirichlet_coefficient`` -- list of Dirichlet coefficients of the
      `L`-function. Only first `M` coefficients are needed if they are periodic.

    - ``period`` -- if the coefficients are periodic, this should be the
      period of the coefficients

    - ``Q`` -- see above

    - ``OMEGA`` -- see above

    - ``kappa`` -- list of the values of `\kappa_j` in the functional equation

    - ``gamma`` -- list of the values of `\gamma_j` in the functional equation

    - ``pole`` -- list of the poles of `L`-function

    - ``residue`` -- list of the residues of the `L`-function

    .. NOTE::

         If an `L`-function satisfies `\Lambda(s) = \omega Q^s \Lambda(k-s)`,
         by replacing `s` by `s+(k-1)/2`, one can get it in the form we need.
    """

    def __init__(self, name, what_type_L, dirichlet_coefficient,
                 period, Q, OMEGA, gamma, lambd, pole, residue):
        r"""
        Initialize an `L`-function with integer coefficients.

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[2] #This is a quadratic character
            sage: L=Lfunction_from_character(chi, type='int')
            sage: type(L)
            <class 'sage.libs.lcalc.lcalc_Lfunction.Lfunction_I'>
        """
        Lfunction.__init__(self, name, what_type_L, dirichlet_coefficient,  period, Q, OMEGA, gamma,lambd, pole,residue)
        self._repr += " with integer Dirichlet coefficients"

    # override
    cdef void _init_fun(self, char *NAME, int what_type, dirichlet_coeff, long long Period, double q,  c_Complex w, int A, double *g, c_Complex *l, int n_poles, c_Complex *p, c_Complex *r) noexcept:
        cdef int N = len(dirichlet_coeff)
        cdef Integer tmpi
        cdef int * coeffs = new_ints(N+1)  # lcalc ignores 0th coefficient
        for i from 0 <= i< N by 1:
            tmpi=Integer(dirichlet_coeff[i])
            coeffs[i+1] = mpz_get_si(tmpi.value)
        self.thisptr=new_c_Lfunction_I(NAME, what_type,  N, coeffs, Period, q,  w,  A, g, l, n_poles, p, r)
        del_ints(coeffs)

    cdef inline c_Complex _value(self, c_Complex s, int derivative) noexcept:
        return (<c_Lfunction_I *>(self.thisptr)).value(s, derivative, "pure")

    cdef inline c_Complex _hardy_z_function(self, c_Complex s) noexcept:
        return (<c_Lfunction_I *>(self.thisptr)).value(s, 0, "rotated pure")

    cdef int _compute_rank(self) noexcept:
        return (<c_Lfunction_I *>(self.thisptr)).compute_rank()

    cdef void _find_zeros_v(self, double T1, double T2, double stepsize, doublevec *result) noexcept:
        (<c_Lfunction_I *>self.thisptr).find_zeros_v(T1,T2,stepsize,result[0])

    cdef double _typedN(self, double T) noexcept:
        return (<c_Lfunction_I *>self.thisptr).N(T)

    cdef int _find_zeros(self, long count, long start, double max_refine, int rank, const char* message_stamp, doublevec *result) noexcept:
        (<c_Lfunction_I *>self.thisptr).find_zeros(count, start, max_refine, rank, message_stamp, result)

    # debug tools
    def _print_data_to_standard_output(self):
        """
        This is used in debugging. It prints out information from
        the C++ object behind the scenes. It will use standard output.

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[2] #This is a quadratic character
            sage: L=Lfunction_from_character(chi, type='int')
            sage: L._print_data_to_standard_output() # tol 1e-8
            -----------------------------------------------
            <BLANKLINE>
            Name of L_function:
            number of dirichlet coefficients = 5
            coefficients are periodic
            b[1] = 1
            b[2] = -1
            b[3] = -1
            b[4] = 1
            b[5] = 0
            <BLANKLINE>
            Q = 1.26156626101
            OMEGA = (1,0)
            a = 1 (the quasi degree)
            gamma[1] =0.5    lambda[1] =(0,0)
            <BLANKLINE>
            <BLANKLINE>
            number of poles (of the completed L function) = 0
            -----------------------------------------------
            <BLANKLINE>
        """
        (<c_Lfunction_I *>self.thisptr).print_data_L()

    def __dealloc__(self):
        """
        Deallocate memory used
        """
        del_c_Lfunction_I(<c_Lfunction_I *>(self.thisptr))

##############################################################################
# Lfunction_D: `L`-functions with double (real) Dirichlet Coefficients
##############################################################################

cdef class Lfunction_D(Lfunction):
    r"""
    The ``Lfunction_D`` class is used to represent `L`-functions
    with real Dirichlet coefficients. We assume that `L`-functions
    satisfy the following functional equation.

    .. MATH::

        \Lambda(s) = \omega Q^s \overline{\Lambda(1-\bar s)}

    where

    .. MATH::

        \Lambda(s) = Q^s \left( \prod_{j=1}^a \Gamma(\kappa_j s + \gamma_j) \right) L(s)

    See (23) in :arxiv:`math/0412181`

    INPUT:

    - ``what_type_L`` -- integer; this should be set to 1 if the coefficients are
      periodic and 0 otherwise

    - ``dirichlet_coefficient`` -- list of Dirichlet coefficients of the
      `L`-function. Only first `M` coefficients are needed if they are periodic.

    - ``period`` -- if the coefficients are periodic, this should be the
      period of the coefficients

    - ``Q`` -- see above

    - ``OMEGA`` -- see above

    - ``kappa`` -- list of the values of `\kappa_j` in the functional equation

    - ``gamma`` -- list of the values of `\gamma_j` in the functional equation

    - ``pole`` -- list of the poles of `L`-function

    - ``residue`` -- list of the residues of the `L`-function

    .. NOTE::

        If an `L`-function satisfies `\Lambda(s) = \omega Q^s \Lambda(k-s)`,
        by replacing `s` by `s+(k-1)/2`, one can get it in the form we need.
    """
    def __init__(self, name, what_type_L, dirichlet_coefficient,
                 period, Q, OMEGA, gamma, lambd, pole, residue):
        r"""
        Initialize an `L`-function with real coefficients.

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[2] #This is a quadratic character
            sage: L=Lfunction_from_character(chi, type='double')
            sage: type(L)
            <class 'sage.libs.lcalc.lcalc_Lfunction.Lfunction_D'>
        """
        Lfunction.__init__(self, name, what_type_L, dirichlet_coefficient,  period, Q, OMEGA, gamma,lambd, pole,residue)
        self._repr += " with real Dirichlet coefficients"

    # override
    cdef void _init_fun(self, char *NAME, int what_type, dirichlet_coeff, long long Period, double q,  c_Complex w, int A, double *g, c_Complex *l, int n_poles, c_Complex *p, c_Complex *r) noexcept:
        cdef int i
        cdef RealNumber tmpr
        cdef int N = len(dirichlet_coeff)
        cdef double * coeffs = new_doubles(N+1)  # lcalc ignores 0th position
        for i in range(N):
            tmpr = RRR(dirichlet_coeff[i])
            coeffs[i+1] = mpfr_get_d(tmpr.value, MPFR_RNDN)
        self.thisptr=new_c_Lfunction_D(NAME, what_type,  N, coeffs, Period, q,  w,  A, g, l, n_poles, p, r)
        del_doubles(coeffs)

    cdef inline c_Complex _value(self, c_Complex s, int derivative) noexcept:
        return (<c_Lfunction_D *>(self.thisptr)).value(s, derivative, "pure")

    cdef inline c_Complex _hardy_z_function(self, c_Complex s) noexcept:
        return (<c_Lfunction_D *>(self.thisptr)).value(s, 0, "rotated pure")

    cdef inline int _compute_rank(self) noexcept:
        return (<c_Lfunction_D *>(self.thisptr)).compute_rank()

    cdef void _find_zeros_v(self, double T1, double T2, double stepsize, doublevec *result) noexcept:
        (<c_Lfunction_D *>self.thisptr).find_zeros_v(T1,T2,stepsize,result[0])

    cdef double _typedN(self, double T) noexcept:
        return (<c_Lfunction_D *>self.thisptr).N(T)

    cdef int _find_zeros(self, long count, long start, double max_refine, int rank, const char* message_stamp, doublevec *result) noexcept:
        (<c_Lfunction_D *>self.thisptr).find_zeros(count, start, max_refine, rank, message_stamp, result)

    # debug tools
    def _print_data_to_standard_output(self):
        """
        This is used in debugging. It prints out information from
        the C++ object behind the scenes. It will use standard output.

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[2] #This is a quadratic character
            sage: L=Lfunction_from_character(chi, type='double')
            sage: L._print_data_to_standard_output() # tol 1e-8
            -----------------------------------------------
            <BLANKLINE>
            Name of L_function:
            number of dirichlet coefficients = 5
            coefficients are periodic
            b[1] = 1
            b[2] = -1
            b[3] = -1
            b[4] = 1
            b[5] = 0
            <BLANKLINE>
            Q = 1.26156626101
            OMEGA = (1,0)
            a = 1 (the quasi degree)
            gamma[1] =0.5    lambda[1] =(0,0)
            <BLANKLINE>
            <BLANKLINE>
            number of poles (of the completed L function) = 0
            -----------------------------------------------
            <BLANKLINE>
        """
        (<c_Lfunction_D *>self.thisptr).print_data_L()

    def __dealloc__(self):
        """
        Deallocate memory used
        """
        del_c_Lfunction_D(<c_Lfunction_D *>(self.thisptr))

##############################################################################
# Lfunction_C: L-functions with Complex Dirichlet Coefficients
##############################################################################

cdef class Lfunction_C:
    r"""
    The ``Lfunction_C`` class is used to represent `L`-functions
    with complex Dirichlet Coefficients. We assume that `L`-functions
    satisfy the following functional equation.

    .. MATH::

        \Lambda(s) = \omega Q^s \overline{\Lambda(1-\bar s)}

    where

    .. MATH::

        \Lambda(s) = Q^s \left( \prod_{j=1}^a \Gamma(\kappa_j s + \gamma_j) \right) L(s)

    See (23) in :arxiv:`math/0412181`

    INPUT:

    - ``what_type_L`` -- integer; this should be set to 1 if the coefficients are
      periodic and 0 otherwise

    - ``dirichlet_coefficient`` -- list of Dirichlet coefficients of the
      `L`-function. Only first `M` coefficients are needed if they are periodic.

    - ``period`` -- if the coefficients are periodic, this should be the
      period of the coefficients

    - ``Q`` -- see above

    - ``OMEGA`` -- see above

    - ``kappa`` -- list of the values of `\kappa_j` in the functional equation

    - ``gamma`` -- list of the values of `\gamma_j` in the functional equation

    - ``pole`` -- list of the poles of `L`-function

    - ``residue`` -- list of the residues of the `L`-function

    .. NOTE::

        If an `L`-function satisfies `\Lambda(s) = \omega Q^s \Lambda(k-s)`,
        by replacing `s` by `s+(k-1)/2`, one can get it in the form we need.
    """

    def __init__(self, name, what_type_L, dirichlet_coefficient,
                 period, Q, OMEGA, gamma, lambd, pole, residue):
        r"""
        Initialize an `L`-function with complex coefficients.

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[1]
            sage: L=Lfunction_from_character(chi, type='complex')
            sage: type(L)
            <class 'sage.libs.lcalc.lcalc_Lfunction.Lfunction_C'>
        """
        Lfunction.__init__(self, name, what_type_L, dirichlet_coefficient,  period, Q, OMEGA, gamma,lambd, pole,residue)
        self._repr += " with complex Dirichlet coefficients"

    # override
    cdef void _init_fun(self, char *NAME, int what_type, dirichlet_coeff, long long Period, double q,  c_Complex w, int A, double *g, c_Complex *l, int n_poles, c_Complex *p, c_Complex *r) noexcept:
        cdef int i
        cdef int N = len(dirichlet_coeff)
        cdef ComplexNumber tmpc

        cdef c_Complex * coeffs = new_Complexes(N+1)
        coeffs[0]=new_Complex(0,0)
        for i from 0 <= i< N by 1:
            tmpc=CCC(dirichlet_coeff[i])
            coeffs[i+1] = new_Complex(mpfr_get_d(tmpc.__re, MPFR_RNDN), mpfr_get_d(tmpc.__im, MPFR_RNDN))

        self.thisptr = new_c_Lfunction_C(NAME, what_type,  N, coeffs, Period, q,  w,  A, g, l, n_poles, p, r)

        del_Complexes(coeffs)

    cdef inline c_Complex _value(self, c_Complex s, int derivative) noexcept:
        return (<c_Lfunction_C *>(self.thisptr)).value(s, derivative, "pure")

    cdef inline c_Complex _hardy_z_function(self, c_Complex s) noexcept:
        return (<c_Lfunction_C *>(self.thisptr)).value(s, 0,"rotated pure")

    cdef inline int _compute_rank(self) noexcept:
        return (<c_Lfunction_C *>(self.thisptr)).compute_rank()

    cdef void _find_zeros_v(self, double T1, double T2, double stepsize, doublevec *result) noexcept:
        (<c_Lfunction_C *>self.thisptr).find_zeros_v(T1,T2,stepsize,result[0])

    cdef double _typedN(self, double T) noexcept:
        return (<c_Lfunction_C *>self.thisptr).N(T)

    cdef int _find_zeros(self, long count, long start, double max_refine, int rank, const char* message_stamp, doublevec *result) noexcept:
        (<c_Lfunction_C *>self.thisptr).find_zeros(count, start, max_refine, rank, message_stamp, result)

    # debug tools
    def _print_data_to_standard_output(self):
        """
        This is used in debugging. It prints out information from
        the C++ object behind the scenes. It will use standard output.

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: chi = DirichletGroup(5)[1]
            sage: L=Lfunction_from_character(chi, type='complex')
            sage: L._print_data_to_standard_output() # tol 1e-8
            -----------------------------------------------
            <BLANKLINE>
            Name of L_function:
            number of dirichlet coefficients = 5
            coefficients are periodic
            b[1] = (1,0)
            b[2] = (0,1)
            b[3] = (0,-1)
            b[4] = (-1,0)
            b[5] = (0,0)
            <BLANKLINE>
            Q = 1.26156626101
            OMEGA = (0.850650808352,0.525731112119)
            a = 1 (the quasi degree)
            gamma[1] =0.5    lambda[1] =(0.5,0)
            <BLANKLINE>
            <BLANKLINE>
            number of poles (of the completed L function) = 0
            -----------------------------------------------
            <BLANKLINE>
        """
        (<c_Lfunction_C *>self.thisptr).print_data_L()

    def __dealloc__(self):
        """
        Deallocate memory used
        """
        del_c_Lfunction_C(<c_Lfunction_C *>(self.thisptr))


##############################################################################
# Zeta function
##############################################################################

cdef class Lfunction_Zeta(Lfunction):
    r"""
    The ``Lfunction_Zeta`` class is used to generate the Riemann zeta function.
    """
    def __init__(self):
        r"""
        Initialize the Riemann zeta function.

        EXAMPLES::

            sage: from sage.libs.lcalc.lcalc_Lfunction import *
            sage: sage.libs.lcalc.lcalc_Lfunction.Lfunction_Zeta()
            The Riemann zeta function
        """
        self.thisptr = new_c_Lfunction_Zeta()
        self._repr = "The Riemann zeta function"

    cdef inline c_Complex _value(self, c_Complex s, int derivative) noexcept:
        return (<c_Lfunction_Zeta *>(self.thisptr)).value(s, derivative, "pure")

    cdef inline c_Complex _hardy_z_function(self, c_Complex s) noexcept:
        return (<c_Lfunction_Zeta *>(self.thisptr)).value(s, 0, "rotated pure")

    cdef inline int _compute_rank(self) noexcept:
        return (<c_Lfunction_Zeta *>(self.thisptr)).compute_rank()

    cdef void _find_zeros_v(self, double T1, double T2, double stepsize, doublevec *result) noexcept:
        (<c_Lfunction_Zeta *>self.thisptr).find_zeros_v(T1,T2,stepsize,result[0])

    cdef double _typedN(self, double T) noexcept:
        return (<c_Lfunction_Zeta *>self.thisptr).N(T)

    cdef int _find_zeros(self, long count, long start, double max_refine, int rank, const char* message_stamp, doublevec *result) noexcept:
        (<c_Lfunction_Zeta *>self.thisptr).find_zeros(count, start, max_refine, rank, message_stamp, result)

    def __dealloc__(self):
        """
        Deallocate memory used
        """
        del_c_Lfunction_Zeta(<c_Lfunction_Zeta *>(self.thisptr))


##############################################################################
# Tools
##############################################################################

def Lfunction_from_character(chi, type='complex'):
    """
    Given a primitive Dirichlet character, this function returns
    an lcalc `L`-function object for the `L`-function of the character.

    INPUT:

    - ``chi`` -- a Dirichlet character
    - ``use_type`` -- string (default: ``'complex'``); type used for the Dirichlet
      coefficients. This can be ``'int'``, ``'double'`` or ``'complex'``.

    OUTPUT: `L`-function object for ``chi``

    EXAMPLES::

        sage: from sage.libs.lcalc.lcalc_Lfunction import Lfunction_from_character
        sage: Lfunction_from_character(DirichletGroup(5)[1])
        L-function with complex Dirichlet coefficients
        sage: Lfunction_from_character(DirichletGroup(5)[2], type='int')
        L-function with integer Dirichlet coefficients
        sage: Lfunction_from_character(DirichletGroup(5)[2], type='double')
        L-function with real Dirichlet coefficients
        sage: Lfunction_from_character(DirichletGroup(5)[1], type='int')
        Traceback (most recent call last):
        ...
        ValueError: For non quadratic characters you must use type="complex"
    """
    if (not chi.is_primitive()):
        raise TypeError("Dirichlet character is not primitive")

    modulus=chi.modulus()
    if chi.is_even():
        a=0
    else:
        a=1

    Q=(RRR(modulus/pi)).sqrt()
    poles=[]
    residues=[]
    period=modulus
    OMEGA=1.0/ ( CCC(0,1)**a * (CCC(modulus)).sqrt()/chi.gauss_sum() )

    if type == "complex":
        dir_coeffs = [CCC(chi(n)) for n in range(1, modulus + 1)]
        return Lfunction_C("", 1,dir_coeffs, period,Q,OMEGA,[.5],[a/2.],poles,residues)
    if type not in ["double", "int"]:
        raise ValueError("unknown type")
    if chi.order() != 2:
        raise ValueError("For non quadratic characters you must use type=\"complex\"")
    if type == "double":
        dir_coeffs = [RRR(chi(n)) for n in range(1, modulus + 1)]
        return Lfunction_D("", 1,dir_coeffs, period,Q,OMEGA,[.5],[a/2.],poles,residues)
    if type == "int":
        dir_coeffs = [Integer(chi(n)) for n in range(1, modulus + 1)]
        return Lfunction_I("", 1,dir_coeffs, period,Q,OMEGA,[.5],[a/2.],poles,residues)


def Lfunction_from_elliptic_curve(E, number_of_coeffs=10000):
    """
    Given an elliptic curve E, return an `L`-function object for
    the function `L(s, E)`.

    INPUT:

    - ``E`` -- an elliptic curve
    - ``number_of_coeffs`` -- integer (default: 10000); the number of
      coefficients to be used when constructing the `L`-function object. Right
      now this is fixed at object creation time, and is not automatically
      set intelligently.

    OUTPUT: `L`-function object for ``L(s, E)``

    EXAMPLES::

        sage: from sage.libs.lcalc.lcalc_Lfunction import Lfunction_from_elliptic_curve
        sage: L = Lfunction_from_elliptic_curve(EllipticCurve('37'))
        sage: L
        L-function with real Dirichlet coefficients
        sage: L.value(0.5).abs() < 1e-8
        True
        sage: (L.value(0.5, derivative=1) - 0.305999773835200).abs() < 1e-6
        True
    """
    Q = RRR(E.conductor()).sqrt() / RRR(2 * pi)
    poles = []
    residues = []
    dir_coeffs = E.anlist(number_of_coeffs)
    dir_coeffs = [RRR(dir_coeffs[i]) / (RRR(i)).sqrt()
                  for i in range(1, number_of_coeffs)]
    OMEGA = E.root_number()
    return Lfunction_D("", 2, dir_coeffs, 0, Q, OMEGA, [1], [.5],
                       poles, residues)



# sage_setup: distribution = sagemath-schemes
# sage.doctest: needs sage.libs.pari
"""
Abelian varieties attached to newforms

TESTS::

    sage: A = AbelianVariety('23a')
    sage: loads(dumps(A)) == A
    True
"""
###########################################################################
#       Copyright (C) 2008 William Stein <wstein@gmail.com>               #
#  Distributed under the terms of the GNU General Public License (GPL)    #
#                  https://www.gnu.org/licenses/                           #
###########################################################################
from sage.misc.lazy_import import lazy_import

from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ

from sage.modular.modform.element import Newform
from sage.modular.arithgroup.congroup_gamma0 import Gamma0_class
from sage.modular.arithgroup.congroup_gamma1 import Gamma1_class
from sage.modular.arithgroup.congroup_gammaH import GammaH_class

from .abvar import ModularAbelianVariety_modsym_abstract
from sage.modular.abvar import homspace
lazy_import('sage.databases.cremona', 'cremona_letter_code')


class ModularAbelianVariety_newform(ModularAbelianVariety_modsym_abstract):
    """
    A modular abelian variety attached to a specific newform.
    """
    def __init__(self, f, internal_name=False):
        """
        Create the modular abelian variety `A_f` attached to the
        newform `f`.

        INPUT:

        - ``f`` -- a newform

        EXAMPLES::

            sage: f = CuspForms(37).newforms('a')[0]
            sage: f.abelian_variety()
            Newform abelian subvariety 37a of dimension 1 of J0(37)

            sage: AbelianVariety(Newforms(1, 12)[0])
            Traceback (most recent call last):
            ...
            TypeError: f must have weight 2
        """
        if not isinstance(f, Newform):
            raise TypeError("f must be a newform")
        if f.weight() != 2:
            raise TypeError("f must have weight 2")
        self.__f = f
        self._is_hecke_stable = True
        K = f.qexp().base_ring()
        if K == QQ:
            variable_name = None
        else:
            variable_name = K.variable_name()
        self.__named_newforms = {variable_name: self.__f}
        if not internal_name:
            self.__named_newforms[None] = self.__f
        ModularAbelianVariety_modsym_abstract.__init__(self, (f.group(),), QQ,
            is_simple=True, newform_level=(f.level(), f.group()),
            isogeny_number=f.number(), number=0)

    def _modular_symbols(self, sign=0):
        """
        EXAMPLES::

            sage: f = CuspForms(52).newforms('a')[0]
            sage: A = f.abelian_variety()
            sage: A._modular_symbols()
            Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 15 for Gamma_0(52) of weight 2 with sign 0 over Rational Field
            sage: A._modular_symbols(1)
            Modular Symbols subspace of dimension 1 of Modular Symbols space of dimension 10 for Gamma_0(52) of weight 2 with sign 1 over Rational Field
            sage: A._modular_symbols(-1)
            Modular Symbols subspace of dimension 1 of Modular Symbols space of dimension 5 for Gamma_0(52) of weight 2 with sign -1 over Rational Field
        """
        return self.__f.modular_symbols(sign=sign)

    def newform(self, names=None):
        r"""
        Return the newform that this modular abelian variety is attached to.

        EXAMPLES::

            sage: f = Newform('37a')
            sage: A = f.abelian_variety()
            sage: A.newform()
            q - 2*q^2 - 3*q^3 + 2*q^4 - 2*q^5 + O(q^6)
            sage: A.newform() is f
            True

        If the a variable name has not been specified, we must specify one::

            sage: A = AbelianVariety('67b')
            sage: A.newform()
            Traceback (most recent call last):
            ...
            TypeError: You must specify the name of the generator.
            sage: A.newform('alpha')
            q + alpha*q^2 + (-alpha - 3)*q^3 + (-3*alpha - 3)*q^4 - 3*q^5 + O(q^6)

        If the eigenform is actually over `\QQ` then we don't have to specify
        the name::

            sage: A = AbelianVariety('67a')
            sage: A.newform()
            q + 2*q^2 - 2*q^3 + 2*q^4 + 2*q^5 + O(q^6)
        """
        try:
            return self.__named_newforms[names]
        except KeyError:
            self.__named_newforms[names] = Newform(self.__f.parent().change_ring(QQ), self.__f.modular_symbols(1), names=names, check=False)
            return self.__named_newforms[names]

    def label(self) -> str:
        """
        Return canonical label that defines this newform modular
        abelian variety.

        OUTPUT: string

        EXAMPLES::

            sage: A = AbelianVariety('43b')
            sage: A.label()
            '43b'
        """
        G = self.__f.group()
        if isinstance(G, Gamma0_class):
            group = ''
        elif isinstance(G, Gamma1_class):
            group = 'G1'
        elif isinstance(G, GammaH_class):
            group = 'GH[' + ','.join(str(z) for z in G._generators_for_H()) + ']'
        return '%s%s%s' % (self.level(), cremona_letter_code(self.factor_number()), group)

    def factor_number(self):
        """
        Return factor number.

        OUTPUT: int

        EXAMPLES::

            sage: A = AbelianVariety('43b')
            sage: A.factor_number()
            1
        """
        try:
            return self.__factor_number
        except AttributeError:
            self.__factor_number = self.__f.number()
            return self.__factor_number

    def _repr_(self) -> str:
        """
        String representation of this modular abelian variety.

        EXAMPLES::

            sage: AbelianVariety('37a')._repr_()
            'Newform abelian subvariety 37a of dimension 1 of J0(37)'
        """
        return "Newform abelian subvariety %s of dimension %s of %s" % (
            self.newform_label(), self.dimension(), self._ambient_repr())

    def endomorphism_ring(self):
        """
        Return the endomorphism ring of this newform abelian variety.

        EXAMPLES::

            sage: A = AbelianVariety('23a')
            sage: E = A.endomorphism_ring(); E
            Endomorphism ring of Newform abelian subvariety 23a of dimension 2 of J0(23)

        We display the matrices of these two basis matrices::

            sage: E.0.matrix()
            [1 0 0 0]
            [0 1 0 0]
            [0 0 1 0]
            [0 0 0 1]
            sage: E.1.matrix()
            [ 0  1 -1  0]
            [ 0  1 -1  1]
            [-1  2 -2  1]
            [-1  1  0 -1]

        The result is cached::

            sage: E is A.endomorphism_ring()
            True
        """
        try:
            return self.__endomorphism_ring
        except AttributeError:
            pass

        E = homspace.EndomorphismSubring(self)
        self.__endomorphism_ring = E
        return self.__endomorphism_ring

    def _calculate_endomorphism_generators(self):
        """
        EXAMPLES::

            sage: A = AbelianVariety('43b')
            sage: B = A.endomorphism_ring(); B   # indirect doctest
            Endomorphism ring of Newform abelian subvariety 43b of dimension 2 of J0(43)
            sage: [b.matrix() for b in B.gens()]
            [
            [1 0 0 0]  [ 0  1  0  0]
            [0 1 0 0]  [ 1 -2  0  0]
            [0 0 1 0]  [ 1  0 -2 -1]
            [0 0 0 1], [ 0  1 -1  0]
            ]
        """
        M = self.modular_symbols()
        bound = M.sturm_bound()

        d = self.dimension()
        T1list = self.hecke_operator(1).matrix().list()
        EndVecZ = ZZ**(len(T1list))
        V = EndVecZ.submodule([T1list])
        n = 2

        while V.dimension() < d:
            W = EndVecZ.submodule([((self.hecke_operator(n).matrix())**i).list()
                                   for i in range(1, d + 1)])
            V = V + W
            n += 1
            if n > bound:
                raise ArithmeticError("Error computing endomorphism generators")

        return V.saturation().basis()



# sage_setup: distribution = sagemath-modules
r"""
Elements of free modules of finite rank

The class :class:`FiniteRankFreeModuleElement` implements elements of
free modules of finite rank over a commutative ring.

AUTHORS:

- Eric Gourgoulhon, Michal Bejger (2014-2015): initial version
- Eric Gourgoulhon (2017): class :class:`FiniteRankFreeModuleElement` inherits
  from :class:`~sage.tensor.modules.alternating_contr_tensor.AlternatingContrTensor`

REFERENCES:

- Chap. 21 of R. Godement : *Algebra* [God1968]_
- Chap. 12 of J. M. Lee: *Introduction to Smooth Manifolds* [Lee2013]_ (only
  when the free module is a vector space)
- Chap. 2 of B. O'Neill: *Semi-Riemannian Geometry* [ONe1983]_
"""

#******************************************************************************
#       Copyright (C) 2015 Eric Gourgoulhon <eric.gourgoulhon@obspm.fr>
#       Copyright (C) 2015 Michal Bejger <bejger@camk.edu.pl>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
#******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from sage.tensor.modules.alternating_contr_tensor import AlternatingContrTensor
from sage.tensor.modules.comp import Components

if TYPE_CHECKING:
    from sage.tensor.modules.finite_rank_free_module import FiniteRankFreeModule
    from sage.tensor.modules.free_module_basis import FreeModuleBasis


class FiniteRankFreeModuleElement(AlternatingContrTensor):
    r"""
    Element of a free module of finite rank over a commutative ring.

    This is a Sage *element* class, the corresponding *parent* class being
    :class:`~sage.tensor.modules.finite_rank_free_module.FiniteRankFreeModule`.

    The class :class:`FiniteRankFreeModuleElement` inherits from
    :class:`~sage.tensor.modules.alternating_contr_tensor.AlternatingContrTensor`
    because the elements of a free module `M` of finite rank over a commutative
    ring `R` are identified with tensors of type `(1,0)` on `M` via the
    canonical map

    .. MATH::

        \begin{array}{lllllll}
        \Phi: & M & \longrightarrow & M^{**}  &  &   & \\
              & v & \longmapsto & \bar v : & M^* & \longrightarrow & R \\
              &   &             &          & a   & \longmapsto & a(v)
        \end{array}

    Note that for free modules of finite rank, this map is actually an
    isomorphism, enabling the canonical identification: `M^{**}= M`.

    INPUT:

    - ``fmodule`` -- free module `M` of finite rank over a commutative ring
      `R`, as an instance of
      :class:`~sage.tensor.modules.finite_rank_free_module.FiniteRankFreeModule`
    - ``name`` -- (default: ``None``) name given to the element
    - ``latex_name`` -- (default: ``None``) LaTeX symbol to denote the element;
      if none is provided, the LaTeX symbol is set to ``name``

    EXAMPLES:

    Let us consider a rank-3 free module `M` over `\ZZ`::

        sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
        sage: e = M.basis('e') ; e
        Basis (e_0,e_1,e_2) on the Rank-3 free module M over the Integer Ring

    There are three ways to construct an element of the free module `M`:
    the first one (recommended) is using the free module::

        sage: v = M([2,0,-1], basis=e, name='v') ; v
        Element v of the Rank-3 free module M over the Integer Ring
        sage: v.display()  # expansion on the default basis (e)
        v = 2 e_0 - e_2
        sage: v.parent() is M
        True

    The second way is to construct a tensor of type `(1,0)` on `M` (cf. the
    canonical identification `M^{**} = M` recalled above)::

        sage: v2 = M.tensor((1,0), name='v')
        sage: v2[0], v2[2] = 2, -1 ; v2
        Element v of the Rank-3 free module M over the Integer Ring
        sage: v2.display()
        v = 2 e_0 - e_2
        sage: v2 == v
        True

    Finally, the third way is via some linear combination of the basis
    elements::

        sage: v3 = 2*e[0] - e[2]
        sage: v3.set_name('v') ; v3 # in this case, the name has to be set separately
        Element v of the Rank-3 free module M over the Integer Ring
        sage: v3.display()
        v = 2 e_0 - e_2
        sage: v3 == v
        True

    The canonical identification `M^{**} = M` is implemented by letting the
    module elements act on linear forms, providing the same result as the
    reverse operation (cf. the map `\Phi` defined above)::

        sage: a = M.linear_form(name='a')
        sage: a[:] = (2, 1, -3) ; a
        Linear form a on the Rank-3 free module M over the Integer Ring
        sage: v(a)
        7
        sage: a(v)
        7
        sage: a(v) == v(a)
        True

    .. RUBRIC:: ARITHMETIC EXAMPLES

    Addition::

        sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
        sage: e = M.basis('e') ; e
        Basis (e_0,e_1,e_2) on the Rank-3 free module M over the Integer Ring
        sage: a = M([0,1,3], name='a') ; a
        Element a of the Rank-3 free module M over the Integer Ring
        sage: a.display()
        a = e_1 + 3 e_2
        sage: b = M([2,-2,1], name='b') ; b
        Element b of the Rank-3 free module M over the Integer Ring
        sage: b.display()
        b = 2 e_0 - 2 e_1 + e_2
        sage: s = a + b ; s
        Element a+b of the Rank-3 free module M over the Integer Ring
        sage: s.display()
        a+b = 2 e_0 - e_1 + 4 e_2
        sage: all(s[i] == a[i] + b[i] for i in M.irange())
        True

    Subtraction::

        sage: s = a - b ; s
        Element a-b of the Rank-3 free module M over the Integer Ring
        sage: s.display()
        a-b = -2 e_0 + 3 e_1 + 2 e_2
        sage: all(s[i] == a[i] - b[i] for i in M.irange())
        True

    Multiplication by a scalar::

        sage: s = 2*a ; s
        Element of the Rank-3 free module M over the Integer Ring
        sage: s.display()
        2 e_1 + 6 e_2
        sage: a.display()
        a = e_1 + 3 e_2

    Tensor product::

        sage: s = a*b ; s
        Type-(2,0) tensor a⊗b on the Rank-3 free module M over the Integer Ring
        sage: s.symmetries()
        no symmetry;  no antisymmetry
        sage: s[:]
        [ 0  0  0]
        [ 2 -2  1]
        [ 6 -6  3]
        sage: s = a*s ; s
        Type-(3,0) tensor a⊗a⊗b on the Rank-3 free module M over the Integer Ring
        sage: s[:]
        [[[0, 0, 0], [0, 0, 0], [0, 0, 0]],
         [[0, 0, 0], [2, -2, 1], [6, -6, 3]],
         [[0, 0, 0], [6, -6, 3], [18, -18, 9]]]

    Exterior product::

        sage: s = a.wedge(b) ; s
        Alternating contravariant tensor a∧b of degree 2 on the Rank-3 free
         module M over the Integer Ring
        sage: s.display()
        a∧b = -2 e_0∧e_1 - 6 e_0∧e_2 + 7 e_1∧e_2
    """

    def __init__(
        self,
        fmodule: FiniteRankFreeModule,
        name: str | None = None,
        latex_name: str | None = None,
    ):
        r"""
        TESTS::

            sage: from sage.tensor.modules.free_module_element import FiniteRankFreeModuleElement
            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: v = FiniteRankFreeModuleElement(M, name='v')
            sage: v[e,:] = (-2, 1, 3)
            sage: TestSuite(v).run(skip='_test_category') # see below

        In the above test suite, _test_category fails because v is not an
        instance of v.parent().category().element_class. Actually module
        elements must be constructed via FiniteRankFreeModule.element_class and
        not by a direct call to FiniteRankFreeModuleElement::

            sage: v1 = M.element_class(M, name='v')
            sage: v1[e,:] = (-2, 1, 3)
            sage: TestSuite(v1).run()
        """
        AlternatingContrTensor.__init__(self, fmodule, 1, name=name,
                                        latex_name=latex_name)

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: M([1,-2,3], name='v')
            Element v of the Rank-3 free module M over the Integer Ring
        """
        description = "Element "
        if self._name is not None:
            description += self._name + " "
        description += "of the {}".format(self._fmodule)
        return description

    def _new_comp(self, basis: FreeModuleBasis) -> Components:
        r"""
        Create some (uninitialized) components of ``self`` in a given basis.

        This method, which is already implemented in
        :meth:`FreeModuleTensor._new_comp`, is redefined here for efficiency.

        INPUT:

        - ``basis`` -- basis of the free module on which ``self`` is defined

        OUTPUT: an instance of :class:`~sage.tensor.modules.comp.Components`

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: v = M([1,-2,3], name='v')
            sage: v._new_comp(e)
            1-index components w.r.t. Basis (e_0,e_1,e_2) on the
             Rank-3 free module M over the Integer Ring
            sage: type(v._new_comp(e))
            <class 'sage.tensor.modules.comp.Components'>
        """
        fmodule = self._fmodule  # the base free module
        return Components(fmodule._ring, basis, 1, start_index=fmodule._sindex,
                          output_formatter=fmodule._output_formatter)

    def _new_instance(self):
        r"""
        Create an instance of the same class as ``self``.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: v = M([1,-2,3], name='v')
            sage: v._new_instance()
            Element of the Rank-3 free module M over the Integer Ring
            sage: v._new_instance().parent() is v.parent()
            True
        """
        return self.__class__(self._fmodule)



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/long_extras.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    size_t z_sizeinbase(slong n, int b) noexcept
    int z_mul_checked(slong * a, slong b, slong c) noexcept
    mp_limb_signed_t z_randtest(flint_rand_t state) noexcept
    mp_limb_signed_t z_randtest_not_zero(flint_rand_t state) noexcept
    mp_limb_signed_t z_randint(flint_rand_t state, mp_limb_t limit) noexcept
    int z_kronecker(slong a, slong n) noexcept



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.numerical.mip
r"""
Delsarte (or linear programming) bounds

This module provides LP upper bounds for the parameters of codes, introduced in
by P. Delsarte in [De1973]_.

The exact LP solver PPL is used by default, ensuring that no
rounding/overflow problems occur.

AUTHORS:

- Dmitrii V. (Dima) Pasechnik (2012-10): initial implementation

- Dmitrii V. (Dima) Pasechnik (2015, 2021): minor fixes

- Charalampos Kokkalis (2021): Eberlein polynomials, general Q matrix codes
"""
# ****************************************************************************
#       Copyright (C) 2012, 2021 Dima Pasechnik <dimpase@gmail.com>
#       Copyright (C) 2021 Charalampos Kokkalis
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


def krawtchouk(n, q, l, x, check=True):
    r"""
    Compute `K^{n,q}_l(x)`, the Krawtchouk (a.k.a. Kravchuk) polynomial.

    See :wikipedia:`Kravchuk_polynomials`.

    It is defined by the generating function

    .. MATH::

        (1+(q-1)z)^{n-x}(1-z)^x=\sum_{l} K^{n,q}_l(x)z^l

    and is equal to

    .. MATH::

        K^{n,q}_l(x)=\sum_{j=0}^l (-1)^j (q-1)^{(l-j)} \binom{x}{j} \binom{n-x}{l-j}.

    INPUT:

    - ``n``, ``q``, ``x`` -- arbitrary numbers

    - ``l`` -- nonnegative integer

    - ``check`` -- check the input for correctness. ``True`` by
      default. Otherwise, pass it as it is. Use ``check=False`` at
      your own risk.

    .. SEEALSO::

        :class:`Symbolic Krawtchouk polynomials
        <sage.functions.orthogonal_polys.Func_krawtchouk>` `\tilde{K}_l(x; n, p)`
        which are related by

        .. MATH::

            (-q)^l K^{n,q^{-1}}_l(x) = \tilde{K}_l(x; n, 1-q).

    EXAMPLES::

        sage: codes.bounds.krawtchouk(24,2,5,4)
        2224
        sage: codes.bounds.krawtchouk(12300,4,5,6)
        567785569973042442072

    TESTS:

    Check that the bug reported on :issue:`19561` is fixed::

        sage: codes.bounds.krawtchouk(3,2,3,3)
        -1
        sage: codes.bounds.krawtchouk(int(3),int(2),int(3),int(3))
        -1
        sage: codes.bounds.krawtchouk(int(3),int(2),int(3),int(3),check=False)
        -1.0
        sage: codes.bounds.krawtchouk(24,2,5,4)
        2224

    Other unusual inputs::

        sage: codes.bounds.krawtchouk(sqrt(5),1-I*sqrt(3),3,55.3).n()                   # needs sage.symbolic
        211295.892797... + 1186.42763...*I
        sage: codes.bounds.krawtchouk(-5/2,7*I,3,-1/10)                                 # needs sage.symbolic
        480053/250*I - 357231/400
        sage: codes.bounds.krawtchouk(1,1,-1,1)
        Traceback (most recent call last):
        ...
        ValueError: l must be a nonnegative integer
        sage: codes.bounds.krawtchouk(1,1,3/2,1)
        Traceback (most recent call last):
        ...
        TypeError: no conversion of this rational to integer
    """
    from sage.arith.misc import binomial
    from sage.arith.srange import srange
    # Use the expression in equation (55) of MacWilliams & Sloane, pg 151
    # We write jth term = some_factor * (j-1)th term
    if check:
        from sage.rings.integer_ring import ZZ
        l0 = ZZ(l)
        if l0 != l or l0 < 0:
            raise ValueError('l must be a nonnegative integer')
        l = l0
    kraw = jth_term = (q-1)**l * binomial(n, l)  # j=0
    for j in srange(1, l+1):
        jth_term *= -q*(l-j+1)*(x-j+1)/((q-1)*j*(n-j+1))
        kraw += jth_term
    return kraw


def eberlein(n, w, k, u, check=True):
    r"""
    Compute `E^{w,n}_k(x)`, the Eberlein polynomial.

    See :wikipedia:`Eberlein_polynomials`.

    It is defined as:

    .. MATH::

        E^{w,n}_k(u)=\sum_{j=0}^k (-1)^j \binom{u}{j} \binom{w-u}{k-j}
        \binom{n-w-u}{k-j},

    INPUT:

    - ``w``, ``k``, ``x`` -- arbitrary numbers

    - ``n`` -- nonnegative integer

    - ``check`` -- check the input for correctness. ``True`` by
      default. Otherwise, pass it as it is. Use ``check=False`` at
      your own risk.

    EXAMPLES::

        sage: codes.bounds.eberlein(24,10,2,6)
        -9

    TESTS:

    check normal inputs (various formats for arguments) ::

        sage: codes.bounds.eberlein(24,10,2,6)
        -9
        sage: codes.bounds.eberlein(int(24),int(10),int(2),int(6))
        -9
        sage: codes.bounds.eberlein(int(24),int(10),int(2),int(6),check=False)
        -9

    unusual inputs ::

        sage: codes.bounds.eberlein(-1,1,1,1)
        Traceback (most recent call last):
        ...
        ValueError: l must be a nonnegative integer
        sage: codes.bounds.eberlein(1,1,3/2,1)
        Traceback (most recent call last):
        ...
        TypeError: either m or x-m must be an integer
    """
    from sage.arith.misc import binomial
    from sage.arith.srange import srange

    if 2 * w > n:
        return eberlein(n, n - w, k, u)

    if check:
        from sage.rings.integer_ring import ZZ
        n0 = ZZ(n)
        if n0 != n or n0 < 0:
            raise ValueError('l must be a nonnegative integer')
        n = n0

    return sum([(-1)**j*binomial(u, j)*binomial(w-u, k-j)*binomial(n-w-u, k-j)
                for j in srange(k + 1)])


def _delsarte_LP_building(n, d, d_star, q, isinteger, solver, maxc=0):
    r"""
    LP builder - common for the two functions; not exported.

    EXAMPLES::

        sage: from sage.coding.delsarte_bounds import _delsarte_LP_building
        sage: _,p=_delsarte_LP_building(7, 3, 0, 2, False, "PPL")
        sage: p.show()
        Maximization:
          x_0 + x_1 + x_2 + x_3 + x_4 + x_5 + x_6 + x_7
        Constraints:
          constraint_0: 1 <= x_0 <= 1
          constraint_1: 0 <= x_1 <= 0
          constraint_2: 0 <= x_2 <= 0
          constraint_3: -7 x_0 - 5 x_1 - 3 x_2 - x_3 + x_4 + 3 x_5 + 5 x_6 + 7 x_7 <= 0
          ...
          constraint_9: - x_0 + x_1 - x_2 + x_3 - x_4 + x_5 - x_6 + x_7 <= 0
        Variables:
          x_0 is a continuous variable (min=0, max=+oo)
          ...
          x_7 is a continuous variable (min=0, max=+oo)
    """
    from sage.numerical.mip import MixedIntegerLinearProgram

    p = MixedIntegerLinearProgram(maximization=True, solver=solver)
    A = p.new_variable(integer=isinteger, nonnegative=True)
    p.set_objective(p.sum([A[r] for r in range(n + 1)]))
    p.add_constraint(A[0] == 1)
    for i in range(1, d):
        p.add_constraint(A[i] == 0)
    for j in range(1, n + 1):
        rhs = p.sum([krawtchouk(n, q, j, r, check=False) * A[r]
                   for r in range(n + 1)])
        if j >= d_star:
            p.add_constraint(0 <= rhs)
        else:  # rhs is proportional to j-th weight of the dual code
            p.add_constraint(0 == rhs)

    if maxc > 0:
        p.add_constraint(p.sum([A[r] for r in range(n + 1)]), max=maxc)
    return A, p


def _delsarte_cwc_LP_building(n, d, w, solver, isinteger):
    r"""
    LP builder for Delsarte's LP for constant weight codes.

    It is used in :func:`delsarte_bound_constant_weight_code`; not exported.

    INPUT:

    - ``n`` -- the code length

    - ``w`` -- the weight of the code

    - ``d`` -- the (lower bound on) minimal distance of the code

    - ``solver`` -- the LP/ILP solver to be used. Defaults to
      ``PPL``. It is arbitrary precision, thus there will be no
      rounding errors. With other solvers (see
      :class:`MixedIntegerLinearProgram` for the list), you are on
      your own!

    - ``isinteger`` -- if ``True``, uses an integer programming solver
      (ILP), rather that an LP solver. Can be very slow if set to
      ``True``.

    EXAMPLES::

        sage: from sage.coding.delsarte_bounds import _delsarte_cwc_LP_building
        sage: _,p=_delsarte_cwc_LP_building(17, 4, 3, "PPL", False)
        sage: p.show()
        Maximization:
          x_0 + x_1 + 1
        Constraints:
          constraint_0: -1 <= 4/21 x_0 - 3/14 x_1
          constraint_1: -1 <= -23/273 x_0 + 3/91 x_1
          constraint_2: -1 <= 1/91 x_0 - 1/364 x_1
        Variables:
          x_0 is a continuous variable (min=0, max=+oo)
          x_1 is a continuous variable (min=0, max=+oo)
    """
    from sage.arith.misc import binomial
    from sage.numerical.mip import MixedIntegerLinearProgram

    p = MixedIntegerLinearProgram(maximization=True, solver=solver)
    A = p.new_variable(integer=isinteger, nonnegative=True)
    p.set_objective(p.sum([A[2*r] for r in range(d//2, w+1)]) + 1)

    def _q(k, i):
        mu_i = 1
        v_i = binomial(w, i)*binomial(n-w, i)
        return mu_i*eberlein(n, w, i, k)/v_i

    for k in range(1, w+1):
        p.add_constraint(p.sum([A[2*i]*_q(k, i) for i in range(d//2, w+1)]),
                         min=-1)

    return A, p


def delsarte_bound_constant_weight_code(n, d, w, return_data=False,
                                        solver='PPL', isinteger=False):
    r"""
    Find the Delsarte bound on a constant weight code.

    Find the Delsarte bound on a constant weight code of weight ``w``, length
    ``n``, lower bound on minimal distance ``d``.

    INPUT:

    - ``n`` -- the code length

    - ``d`` -- the (lower bound on) minimal distance of the code

    - ``w`` -- the weight of the code

    - ``return_data`` -- if ``True``, return a triple
      ``(W,LP,bound)``, where ``W`` is a weights vector, and ``LP``
      the Delsarte upper bound LP; both of them are Sage LP data.
      ``W`` need not be a weight distribution of a code.

    - ``solver`` -- the LP/ILP solver to be used. Defaults to
      ``PPL``. It is arbitrary precision, thus there will be no
      rounding errors. With other solvers (see
      :class:`MixedIntegerLinearProgram` for the list), you are on
      your own!

    - ``isinteger`` -- if ``True``, uses an integer programming solver
      (ILP), rather that an LP solver. Can be very slow if set to
      ``True``.

    EXAMPLES:

    The bound on the size of codes of length 17, weight 3, and minimal distance 4::

       sage: codes.bounds.delsarte_bound_constant_weight_code(17, 4, 3)
       45
       sage: a, p, val = codes.bounds.delsarte_bound_constant_weight_code(17, 4, 3, return_data=True)
       sage: [j for i,j in p.get_values(a).items()]
       [21, 70/3]

    The stricter bound (using ILP) on codes of length 17, weight 3, and minimal
    distance 4::

       sage: codes.bounds.delsarte_bound_constant_weight_code(17, 4, 3, isinteger=True)
       43
    """
    from sage.numerical.mip import MIPSolverException

    if d < 4:
        raise ValueError("Violated constraint d>=4 for "
                         "Binary Constant Weight Codes")

    if d >= 2*w or 2*w > n:
        raise ValueError("Violated constraint d<2w<=n for "
                         "Binary Constant Weight Codes")

    # minimum distance is even => if there is an odd lower bound on d we can
    # increase it by 1
    if d % 2:
        d += 1

    A, p = _delsarte_cwc_LP_building(n, d, w, solver, isinteger)
    try:
        bd = p.solve()
    except MIPSolverException as exc:
        print(f"Solver exception: {exc}")
        return (A, p, False) if return_data else False

    return (A, p, bd) if return_data else int(bd)


def delsarte_bound_hamming_space(n, d, q, return_data=False,
                                 solver='PPL', isinteger=False):
    r"""
    Find the Delsarte bound on codes in ``H_q^n`` of minimal distance ``d``.

    Find the Delsarte bound [De1973]_ on the size of codes in
    the Hamming space ``H_q^n`` of minimal distance ``d``.

    INPUT:

    - ``n`` -- the code length

    - ``d`` -- the (lower bound on) minimal distance of the code

    - ``q`` -- the size of the alphabet

    - ``return_data`` -- if ``True``, return a triple
      ``(W,LP,bound)``, where ``W`` is a weights vector, and ``LP``
      the Delsarte upper bound LP; both of them are Sage LP data.
      ``W`` need not be a weight distribution of a code.

    - ``solver`` -- the LP/ILP solver to be used. Defaults to
      ``PPL``. It is arbitrary precision, thus there will be no
      rounding errors. With other solvers (see
      :class:`MixedIntegerLinearProgram` for the list), you are on
      your own!

    - ``isinteger`` -- if ``True``, uses an integer programming solver
      (ILP), rather that an LP solver. Can be very slow if set to
      ``True``.

    EXAMPLES:

    The bound on the size of the `\GF{2}`-codes of length 11 and minimal distance 6::

       sage: codes.bounds.delsarte_bound_hamming_space(11, 6, 2)
       12
       sage: a, p, val = codes.bounds.delsarte_bound_hamming_space(11, 6, 2, return_data=True)
       sage: [j for i,j in p.get_values(a).items()]
       [1, 0, 0, 0, 0, 0, 11, 0, 0, 0, 0, 0]

    The bound on the size of the `\GF{2}`-codes of length 24 and minimal distance
    8, i.e. parameters of the extended binary Golay code::

       sage: a,p,x = codes.bounds.delsarte_bound_hamming_space(24,8,2,return_data=True)
       sage: x
       4096
       sage: [j for i,j in p.get_values(a).items()]
       [1, 0, 0, 0, 0, 0, 0, 0, 759, 0, 0, 0, 2576, 0, 0, 0, 759, 0, 0, 0, 0, 0, 0, 0, 1]

    The bound on the size of `\GF{4}`-codes of length 11 and minimal distance 3::

       sage: codes.bounds.delsarte_bound_hamming_space(11,3,4)
       327680/3

    An improvement of a known upper bound (150) from https://www.win.tue.nl/~aeb/codes/binary-1.html ::

       sage: a,p,x = codes.bounds.delsarte_bound_hamming_space(23,10,2,return_data=True,isinteger=True); x # long time
       148
       sage: [j for i,j in p.get_values(a).items()]                                      # long time
       [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 95, 0, 2, 0, 36, 0, 14, 0, 0, 0, 0, 0, 0, 0]

    Note that a usual LP, without integer variables, won't do the trick ::

       sage: codes.bounds.delsarte_bound_hamming_space(23,10,2).n(20)
       151.86

    Such an input is invalid::

       sage: codes.bounds.delsarte_bound_hamming_space(11,3,-4)
       Solver exception: PPL : There is no feasible solution
       False
    """
    from sage.numerical.mip import MIPSolverException
    A, p = _delsarte_LP_building(n, d, 0, q, isinteger, solver)
    try:
        bd = p.solve()
    except MIPSolverException as exc:
        print(f"Solver exception: {exc}")
        return (A, p, False) if return_data else False

    return (A, p, bd) if return_data else bd


def delsarte_bound_additive_hamming_space(n, d, q, d_star=1, q_base=0, return_data=False,
                                          solver='PPL', isinteger=False):
    r"""
    Find a modified Delsarte bound on additive codes in Hamming space `H_q^n` of minimal distance `d`.

    Find the Delsarte LP bound on ``F_{q_base}``-dimension of additive
    codes in Hamming space `H_q^n` of minimal distance ``d`` with
    minimal distance of the dual code at least ``d_star``.  If
    ``q_base`` is set to nonzero, then ``q`` is a power of
    ``q_base``, and the code is, formally, linear over
    ``F_{q_base}``. Otherwise it is assumed that ``q_base==q``.

    INPUT:

    - ``n`` -- the code length

    - ``d`` -- the (lower bound on) minimal distance of the code

    - ``q`` -- the size of the alphabet

    - ``d_star`` -- the (lower bound on) minimal distance of the dual code;
      only makes sense for additive codes

    - ``q_base`` -- if ``0``, the code is assumed to be linear. Otherwise,
      ``q=q_base^m`` and the code is linear over ``F_{q_base}``

    - ``return_data`` -- if ``True``, return a triple ``(W,LP,bound)``,
      where ``W`` is a weights vector, and ``LP`` the Delsarte bound
      LP; both of them are Sage LP data.  ``W`` need not be a weight
      distribution of a code, or, if ``isinteger==False``, even have
      integer entries.

    - ``solver`` -- the LP/ILP solver to be used. Defaults to ``'PPL'``. It is
      arbitrary precision, thus there will be no rounding errors. With
      other solvers (see :class:`MixedIntegerLinearProgram` for the
      list), you are on your own!

    - ``isinteger`` -- if ``True``, uses an integer programming solver (ILP),
      rather that an LP solver (can be very slow if set to ``True``)

    EXAMPLES:

    The bound on dimension of linear `\GF{2}`-codes of length 11 and minimal distance 6::

        sage: codes.bounds.delsarte_bound_additive_hamming_space(11, 6, 2)
        3
        sage: a,p,val = codes.bounds.delsarte_bound_additive_hamming_space(\
        ....:                11, 6, 2, return_data=True)
        sage: [j for i,j in p.get_values(a).items()]
        [1, 0, 0, 0, 0, 0, 5, 2, 0, 0, 0, 0]

    The bound on the dimension of linear `\GF{4}`-codes of length 11 and minimal distance 3::

        sage: codes.bounds.delsarte_bound_additive_hamming_space(11,3,4)
        8

    The bound on the `\GF{2}`-dimension of additive `\GF{4}`-codes of length 11 and minimal
    distance 3::

        sage: codes.bounds.delsarte_bound_additive_hamming_space(11,3,4,q_base=2)
        16

    Such a ``d_star`` is not possible::

        sage: codes.bounds.delsarte_bound_additive_hamming_space(11,3,4,d_star=9)
        Solver exception: PPL : There is no feasible solution
        False

    TESTS::

        sage: a,p,x = codes.bounds.delsarte_bound_additive_hamming_space(\
        ....:            19,15,7,return_data=True,isinteger=True)
        sage: [j for i,j in p.get_values(a).items()]
        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 307, 0, 0, 1, 34]
        sage: codes.bounds.delsarte_bound_additive_hamming_space(19,15,7,solver='glpk')
        3
        sage: codes.bounds.delsarte_bound_additive_hamming_space(\
        ....:    19,15,7, isinteger=True, solver='glpk')
        3
    """
    from sage.numerical.mip import MIPSolverException
    if q_base == 0:
        q_base = q

    kk = 0
    while q_base**kk < q:
        kk += 1

    if q_base**kk != q:
        print("Wrong q_base=", q_base, " for q=", q, kk)
        return False

    # this implementation assumes that our LP solver to be unable to do a hot
    # restart with an adjusted constraint

    m = kk*n  # this is to emulate repeat/until block
    bd = q**n+1

    while q_base**m < bd:
        # need to solve the LP repeatedly, as this is a new constraint!
        # we might become infeasible. More precisely, after rounding down
        # to the closest value of q_base^m, the LP, with the constraint that
        # the objective function is at most q_base^m,
        A, p = _delsarte_LP_building(n, d, d_star, q, isinteger,
                                     solver, q_base**m)
        try:
            bd = p.solve()
        except MIPSolverException as exc:
            print("Solver exception:", exc)
            return (A, p, False) if return_data else False
    # rounding the bound down to the nearest power of q_base, for q=q_base^m
    # bd_r = roundres(log(bd, base=q_base))
        m = -1
        while q_base**(m+1) < bd:
            m += 1
        if q_base**(m+1) == bd:
            m += 1

    return (A, p, m) if return_data else m


def _delsarte_Q_LP_building(q, d, solver, isinteger):
    r"""
    LP builder for Delsarte's LP for codes given Q matrix.

    LP builder for Delsarte's LP for codes, given Q matrix;
    used in :func:`delsarte_bound_Q_matrix`; not exported.

    INPUT:

    - ``q`` -- the Q matrix

    - ``d`` -- the (lower bound on) minimal distance of the code

    - ``solver`` -- the LP/ILP solver to be used. Defaults to
      ``PPL``. It is arbitrary precision, thus there will be no
      rounding errors. With other solvers (see
      :class:`MixedIntegerLinearProgram` for the list), you are on
      your own!

    - ``isinteger`` -- if ``True``, uses an integer programming solver
      (ILP), rather that an LP solver. Can be very slow if set to
      ``True``.

    EXAMPLES::

        sage: from sage.coding.delsarte_bounds import _delsarte_Q_LP_building
        sage: q = Matrix([[codes.bounds.krawtchouk(6,2,i,j) for j in range(7)] for i in range(7)])
        sage: _, p = _delsarte_Q_LP_building(q, 2, "PPL", False)
        sage: p.show()
        Maximization:
          x_0 + x_1 + x_2 + x_3 + x_4 + x_5 + x_6
        <BLANKLINE>
        Constraints:
          constraint_0: 1 <= x_0 <= 1
          constraint_1: 0 <= x_1 <= 0
          constraint_2: 0 <= 6 x_0 + 4 x_1 + 2 x_2 - 2 x_4 - 4 x_5 - 6 x_6
          constraint_3: 0 <= 15 x_0 + 5 x_1 - x_2 - 3 x_3 - x_4 + 5 x_5 + 15 x_6
          constraint_4: 0 <= 20 x_0 - 4 x_2 + 4 x_4 - 20 x_6
          constraint_5: 0 <= 15 x_0 - 5 x_1 - x_2 + 3 x_3 - x_4 - 5 x_5 + 15 x_6
          constraint_6: 0 <= 6 x_0 - 4 x_1 + 2 x_2 - 2 x_4 + 4 x_5 - 6 x_6
          constraint_7: 0 <= x_0 - x_1 + x_2 - x_3 + x_4 - x_5 + x_6
        Variables:
          x_0 is a continuous variable (min=0, max=+oo)
          x_1 is a continuous variable (min=0, max=+oo)
          x_2 is a continuous variable (min=0, max=+oo)
          x_3 is a continuous variable (min=0, max=+oo)
          x_4 is a continuous variable (min=0, max=+oo)
          x_5 is a continuous variable (min=0, max=+oo)
          x_6 is a continuous variable (min=0, max=+oo)
    """
    from sage.numerical.mip import MixedIntegerLinearProgram

    n, _ = q.dimensions()  # Q is a square matrix

    p = MixedIntegerLinearProgram(maximization=True, solver=solver)
    A = p.new_variable(integer=isinteger, nonnegative=True)
    p.set_objective(p.sum([A[i] for i in range(n)]))

    p.add_constraint(A[0] == 1)

    try:
        for i in range(1, d):
            p.add_constraint(A[i] == 0)
    except TypeError:
        for i in d:
            p.add_constraint(A[i] == 0)

    for k in range(1, n):
        p.add_constraint(p.sum([q[k][i] * A[i] for i in range(n)]), min=0)

    return A, p


def delsarte_bound_Q_matrix(q, d, return_data=False,
                            solver='PPL', isinteger=False):
    r"""
    Delsarte bound on a code with Q matrix ``q`` and lower bound on min. dist. ``d``.

    Find the Delsarte bound on a code with Q matrix ``q`` and lower bound on
    minimal distance ``d``.

    INPUT:

    - ``q`` -- the Q matrix

    - ``d`` -- the (lower bound on) minimal distance of the code

    - ``return_data`` -- if ``True``, return a triple
      ``(W,LP,bound)``, where ``W`` is a weights vector, and ``LP``
      the Delsarte upper bound LP; both of them are Sage LP data.
      ``W`` need not be a weight distribution of a code.

    - ``solver`` -- the LP/ILP solver to be used. Defaults to
      ``PPL``. It is arbitrary precision, thus there will be no
      rounding errors. With other solvers (see
      :class:`MixedIntegerLinearProgram` for the list), you are on
      your own!

    - ``isinteger`` -- if ``True``, uses an integer programming solver
      (ILP), rather that an LP solver. Can be very slow if set to
      ``True``.

    EXAMPLES:

    The bound on dimension of linear `\GF{2}`-codes of length 10 and minimal distance 6::

        sage: q_matrix = Matrix([[codes.bounds.krawtchouk(10,2,i,j) for i in range(11)]
        ....:                    for j in range(11)])
        sage: codes.bounds.delsarte_bound_Q_matrix(q_matrix, 6)
        2

        sage: a,p,val = codes.bounds.delsarte_bound_Q_matrix(q_matrix, 6, return_data=True)
        sage: [j for i,j in p.get_values(a).items()]
        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]

    TESTS:

    Cases for using Hamming scheme Q matrix::

        sage: q_matrix = Matrix([[codes.bounds.krawtchouk(10,2,i,j) for i in range(11)] for j in range(11)])
        sage: codes.bounds.delsarte_bound_Q_matrix(q_matrix, 6)
        2

        sage: a,p,val = codes.bounds.delsarte_bound_Q_matrix(q_matrix, 6, return_data=True)
        sage: [j for i,j in p.get_values(a).items()]
        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]
    """
    from sage.numerical.mip import MIPSolverException
    from sage.structure.element import Matrix

    if not isinstance(q, Matrix):
        raise ValueError("Input to delsarte_bound_Q_matrix "
                         "should be a sage Matrix()")

    A, p = _delsarte_Q_LP_building(q, d, solver, isinteger)
    try:
        bd = p.solve()
    except MIPSolverException as exc:
        print(f"Solver exception: {exc}")
        return (A, p, False) if return_data else False

    return (A, p, bd) if return_data else bd



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.libs.singular
r"""
Cohomology of coherent sheaves

This module implements Maruyama's method for computing the cohomology of
coherent sheaves on projective schemes. This module is internal, and is not
intended for direct use. Use :meth:`cohomology` method of coherent sheaves on
projective schemes.

Maruyama's method is explained and proved in detail in [Kudo2017]_. Here we
summary the main results necessary to understand the implementation.

Let `M` be a graded module finitely generated over the homogeneous coordinate
ring `S` of the projective `r`-space over a field `k`. We aim for computing the
cohomology groups `H^q(\tilde M)` for the coherent sheaf `\tilde M`.

Let `S=k[x_0,x_2,\dots,x_r]`. Then `M` is a quotient of the free module
`\bigoplus_{i=1}^{t}S` by a submodule. Let

.. MATH::

    0\to\bigoplus_{j=1}^{t_{r+1}}S(-m^{(r+1)}_j)\overset{f_{r+1}}{\longrightarrow}\dots
    \overset{f_1}{\longrightarrow}\bigoplus_{j=1}^{t_0}S(-m^{(0)}_j)\overset{f_0}{\longrightarrow}M\to 0

be a minimal free resolution of `M`. Then it induces a complex of (top) cohomology groups

.. MATH::

    \bigoplus_{j=1}^{t_{i+1}}H^r(\OO_{\PP^r}(-m^{(i+1)}_j))\overset{H^r(f_{i+1})}{\longrightarrow}
    \bigoplus_{j=1}^{t_i}H^r(\OO_{\PP^r}(-m^{(i)}_j))\overset{H^r(f_{i})}{\longrightarrow}
    \bigoplus_{j=1}^{t_{i-1}}H^r(\OO_{\PP^r}(-m^{(i-1)}_j)),

where `i` runs from `1` to `r`. Now it holds that

.. MATH::

    H^q(\tilde M)\cong \ker H^r(f_{r-q})/\im H^r(f_{r-q+1})

for `1\le q\le r - 1` and

.. MATH::

    H^r(\tilde M)\cong \bigoplus_{j=1}^{t_0}H^r(\OO_{\PP^r}(-m^{(0)}_j))/\im H^r(f_1)

and `\dim H^0(\tilde M)` can be computed by the formula

.. MATH::

    \begin{split}
    &\dim \bigoplus_{j=1}^{t_{0}}H^0(\OO_{\PP^r}(-m^{(0)}_j))
    -\dim \bigoplus_{j=1}^{t_{r+1}}H^r(\OO_{\PP^r}(-m^{(r+1)}_j))
    +\dim \bigoplus_{j=1}^{t_{r}}H^r(\OO_{\PP^r}(-m^{(r)}_j)) \\
    &\quad -\rank H^0(f_1)-\rank H^r(f_r)
    \end{split}

in which the complex of (bottom) cohomology groups

.. MATH::

    \bigoplus_{j=1}^{t_{i+1}}H^0(\OO_{\PP^r}(-m^{(i+1)}_j))\overset{H^0(f_{i+1})}{\longrightarrow}
    \bigoplus_{j=1}^{t_i}H^0(\OO_{\PP^r}(-m^{(i)}_j))\overset{H^0(f_{i})}{\longrightarrow}
    \bigoplus_{j=1}^{t_{i-1}}H^0(\OO_{\PP^r}(-m^{(i-1)}_j)),

where `i` runs from `1` to `r` is used.

The implemented algorithm works more generally for twisted coherent sheaves and
accepts as input shifted graded module `M(-n)` with shift `n`.

EXAMPLES:

We define the Fermat cubic surface (a curve in `\PP^2`) and compute its cohomology groups::

    sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
    sage: X = P2.subscheme([x^4 + y^4 + z^4])
    sage: sh = X.structure_sheaf()
    sage: sh.cohomology(1)
    3

Internally the cohomology is computed by Maruyama's method::

    sage: sh._cohomology
    Maruyama Method using S(0) <-- S(-4) <-- 0
    sage: sh._cohomology.H(1)
    Vector space quotient V/W of dimension 3 over Rational Field where
    V: Vector space of degree 3 and dimension 3 over Rational Field
    Basis matrix:
    [1 0 0]
    [0 1 0]
    [0 0 1]
    W: Vector space of degree 3 and dimension 0 over Rational Field
    Basis matrix:
    []
    sage: sh._cohomology.H(1).dimension() == sh.cohomology(1)
    True
    sage: sh.cohomology(2)
    0
    sage: sh._cohomology.H(2)
    Vector space quotient V/W of dimension 0 over Rational Field where
    V: Vector space of dimension 0 over Rational Field
    W: Vector space of degree 0 and dimension 0 over Rational Field
    Basis matrix:
    []
    sage: sh._cohomology.H(2).dimension() == sh.cohomology(2)
    True

The rather complicated form (as a quotient of vector spaces) of the `r`-th
cohomology group ``H(r)`` reflects the internal representation of the
cohomology group in terms of the cohomology groups of twisted structure sheaves
of a projective space.

On the other hand, it is not clear how to represent `H^0(\tilde M)` in terms of
twisted structure sheaves of a projective space. Hence it is merely created
as a vector space over `k` with the correct dimension::

    sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
    sage: X = P2.subscheme([x^4 + y^4 + z^4])
    sage: sh = X.structure_sheaf()
    sage: sh.cohomology(0)
    1
    sage: sh._cohomology.H(0)
    Vector space of dimension 1 over Rational Field
"""

from sage.structure.sage_object import SageObject
from sage.misc.flatten import flatten
from sage.combinat.integer_lists.invlex import IntegerListsLex
from sage.modules.free_module import VectorSpace
from sage.modules.free_module_element import vector
from sage.rings.integer import Integer


class CohomologyGroupBottom(SageObject):
    r"""
    Bottom cohomology group of the twisted structure sheaf of a projective space.

    INPUT:

    - ``S`` -- a direct sum of copies of the coordinate ring of a projective space

    - ``shifts`` -- shifts of the component rings of ``S``

    This represents `\bigoplus_{j=1}^{t_i}H^0(\OO_{\PP^r}(-m^{(i)}_j))` for shifts `m^{(i)}_j`.

    EXAMPLES::

        sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
        sage: X = P2.subscheme([x^4 + y^4 + z^4])
        sage: c = X.structure_sheaf()._cohomology
        sage: c.cohomology_group_bottom(0)
        Bottom Cohomology Group of dimension 1
    """
    def __init__(self, S, shifts):
        """
        Initialize.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: c = X.structure_sheaf(twist=1)._cohomology
            sage: c.cohomology_group_bottom(0)
            Bottom Cohomology Group of dimension 3
        """
        self.graded_ring = S
        self.shifts = shifts

        n = S.ngens()

        basis = []
        summands_basis = []
        summands_index = []
        rank = 0
        for m in self.shifts:
            # list of integer vectors whose entries are all non-negative integers and sum to -m
            l = [vector(e) for e in IntegerListsLex(length=n, min_sum=-m, max_sum=-m)]
            basis += l
            summands_basis.append(l)
            summands_index.append(rank)
            rank += len(l)

        self.summands_basis = summands_basis
        self.summands_index = summands_index
        self.basis = basis
        self.vector_space = VectorSpace(S.base_ring(), rank)
        self.rank = rank

    def __repr__(self):
        r"""
        Return the string representation of ``self``.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: c = X.structure_sheaf()._cohomology
            sage: c.cohomology_group_bottom(0)
            Bottom Cohomology Group of dimension 1
        """
        return f'Bottom Cohomology Group of dimension {self.rank}'


class CohomologyGroupTop(SageObject):
    r"""
    Top cohomology group of the twisted structure sheaf of a projective space.

    INPUT:

    - ``S`` -- a direct sum of copies of the coordinate ring of a projective space

    - ``shifts`` -- shifts of the component rings of ``S``

    This represents `\bigoplus_{j=1}^{t_i}H^r(\OO_{\PP^r}(-m^{(i)}_j))` for shifts `m^{(i)}_j`.

    EXAMPLES::

        sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
        sage: X = P2.subscheme([x^4 + y^4 + z^4])
        sage: c = X.structure_sheaf()._cohomology
        sage: c.cohomology_group_top(0)
        Top Cohomology Group of dimension 0
    """
    def __init__(self, S, shifts):
        """
        Initialize.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: c = X.structure_sheaf(twist=1)._cohomology
            sage: c.cohomology_group_top(1)
            Top Cohomology Group of dimension 1
        """
        self.graded_ring = S
        self.shifts = shifts

        n = S.ngens()

        basis = []
        summands_basis = []
        summands_index = []
        rank = 0
        for m in self.shifts:
            # list of integer vectors whose entries are all negative integers and sum to -m
            l = [-vector(e) for e in IntegerListsLex(length=n, min_sum=m, max_sum=m, min_part=1)]
            basis += l
            summands_basis.append(l)
            summands_index.append(rank)
            rank += len(l)

        self.summands_basis = summands_basis
        self.summands_index = summands_index
        self.basis = basis
        self.vector_space = VectorSpace(S.base_ring(), rank)
        self.rank = rank

    def __repr__(self):
        r"""
        Return the string representation.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: c = X.structure_sheaf()._cohomology
            sage: c.cohomology_group_top(1)
            Top Cohomology Group of dimension 3
        """
        return f'Top Cohomology Group of dimension {self.rank}'


class MaruyamaMethod(SageObject):
    r"""
    This class implements Maruyama's method to compute the cohomology group
    `H^q(\tilde M(n))` as a vector space over the base field `k` and
    `h^q(\tilde M(n))=\dim_kH^q(\tilde M(n))` where `n` denotes the twist.

    INPUT:

    - ``M`` -- a quotient of a free module over `S` by a submodule, where `S`
      is a multi-variate polynomial ring

    - ``twist`` -- (default: 0) an integer

    This class provides :meth:`H` and :meth:`h` as public interface, and all
    other methods are internal.

    EXAMPLES::

        sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
        sage: X = P2.subscheme([x^4 + y^4 + z^4])
        sage: sh = X.structure_sheaf(1)  # twisted sheaf
        sage: sh._cohomology
        Maruyama Method using S(1) <-- S(-3) <-- 0
    """
    def __init__(self, M, twist=0):
        """
        Initialize.

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: sh = X.structure_sheaf(1)  # twisted sheaf
            sage: c = sh._cohomology
            sage: TestSuite(c).run(skip=['_test_pickling'])
        """
        shifts = [-twist for i in range(M.cover().degree())]
        self.resolution = M.relations().graded_free_resolution(shifts=shifts)
        self.base_ring = self.resolution.target().base_ring()
        self.coefficient_field = self.base_ring.base_ring()
        self.projective_space_dimension = self.base_ring.ngens() - 1

    def __repr__(self):
        """
        Return the string representation.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: X.structure_sheaf()._cohomology
            Maruyama Method using S(0) <-- S(-4) <-- 0
        """
        return f'Maruyama Method using {self.resolution}'

    def cohomology_group_bottom(self, i):
        r"""
        Return `i`-th bottom cohomology group `\bigoplus_{j=1}^{t_i}H^0(\OO_{\PP^r}(-m^{(i)}_j))`

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: c = X.structure_sheaf()._cohomology
            sage: c.cohomology_group_bottom(1)
            Bottom Cohomology Group of dimension 0
        """
        S = self.base_ring
        shifts = self.resolution.shifts(i)
        return CohomologyGroupBottom(S, shifts)

    def cohomology_group_top(self, i):
        r"""
        Return `i`-th top cohomology group `\bigoplus_{j=1}^{t_i}H^r(\OO_{\PP^r}(-m^{(i)}_j))`

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: c = X.structure_sheaf()._cohomology
            sage: c.cohomology_group_top(1)
            Top Cohomology Group of dimension 3
        """
        S = self.base_ring
        shifts = self.resolution.shifts(i)
        return CohomologyGroupTop(S, shifts)

    def differential_bottom(self, i):
        r"""
        Return the `i`-th bottom differential map `H^0(f_i)`.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: c = X.structure_sheaf()._cohomology
            sage: c.differential_bottom(1)
            Vector space morphism represented as left-multiplication by the matrix:
            []
            Domain: Vector space of dimension 0 over Rational Field
            Codomain: Vector space of dimension 1 over Rational Field
        """
        H1 = self.cohomology_group_bottom(i)
        H0 = self.cohomology_group_bottom(i - 1)
        M = self.resolution.differential(i).matrix()
        K = self.coefficient_field
        zero = K.zero()

        assert M.ncols() == len(H1.summands_basis)
        assert M.nrows() == len(H0.summands_basis)

        A = []
        for i in range(M.ncols()):
            basis = H1.summands_basis[i]
            for v in basis:
                image = [zero for e in range(H0.rank)]
                for j in range(M.nrows()):
                    f = M[j,i]
                    basis = H0.summands_basis[j]
                    for c, m in zip(f.coefficients(), f.exponents()):
                        u = v + vector(m)
                        assert (sum(u) == -H0.shifts[j])
                        if any(e < 0 for e in u):
                            continue
                        k = H0.summands_index[j] + basis.index(u)
                        image[k] += c
                A.append(vector(K, image))

        return H1.vector_space.hom(A, codomain=H0.vector_space, side='right')

    def differential_top(self, i):
        r"""
        Return the `i`-th top differential map `H^r(f_i)`.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: c = X.structure_sheaf()._cohomology
            sage: c.differential_top(1)
            Vector space morphism represented as left-multiplication by the matrix:
            []
            Domain: Vector space of dimension 3 over Rational Field
            Codomain: Vector space of dimension 0 over Rational Field
        """
        H1 = self.cohomology_group_top(i)
        H0 = self.cohomology_group_top(i - 1)
        M = self.resolution.differential(i).matrix()
        K = self.coefficient_field
        zero = K.zero()

        assert M.ncols() == len(H1.summands_basis)
        assert M.nrows() == len(H0.summands_basis)

        A = []
        for i in range(M.ncols()):
            basis = H1.summands_basis[i]
            for v in basis:
                image = [zero for e in range(H0.rank)]
                for j in range(M.nrows()):
                    f = M[j,i]
                    basis = H0.summands_basis[j]
                    for c, m in zip(f.coefficients(), f.exponents()):
                        u = v + vector(m)
                        assert (sum(u) == -H0.shifts[j])
                        if any(e >= 0 for e in u):
                            continue
                        k = H0.summands_index[j] + basis.index(u)
                        image[k] += c
                A.append(vector(K, image))

        return H1.vector_space.hom(A, codomain=H0.vector_space, side='right')

    def H(self, q):
        r"""
        Return the `q`-th cohomology group `H^q(\tilde M)`.

        INPUT:

        - ``q`` -- non-negative integer

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: c = X.structure_sheaf()._cohomology
            sage: c.H(1)
            Vector space quotient V/W of dimension 3 over Rational Field where
            V: Vector space of degree 3 and dimension 3 over Rational Field
            Basis matrix:
            [1 0 0]
            [0 1 0]
            [0 0 1]
            W: Vector space of degree 3 and dimension 0 over Rational Field
            Basis matrix:
            []
        """
        r = self.projective_space_dimension
        if q == r:
            return self.cohomology_group_top(0).vector_space.quotient(self.differential_top(1).image())
        elif 1 <= q and q < r:
            return self.differential_top(r - q).kernel().quotient(self.differential_top(r - q + 1).image())
        elif q == 0:
            # 0th cohomology group of (global sections) is not realized in
            # terms of cohomology groups and differential maps of twisted
            # structure sheaves of the projective space.
            return VectorSpace(self.coefficient_field, self.h(0))
        elif q > r:
            return VectorSpace(self.coefficient_field, 0)
        raise IndexError('index must be non-negative')

    def h(self, q):
        r"""
        Return the dimension `h^q(\tilde M)` of the `q`-th cohomology group `H^q(\tilde M)`.

        INPUT:

        - ``q`` -- non-negative integer

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme([x^4 + y^4 + z^4])
            sage: c = X.structure_sheaf()._cohomology
            sage: c.h(1)
            3
        """
        r = self.projective_space_dimension
        if q == r:
            return self.cohomology_group_top(0).rank - self.differential_top(1).rank()
        elif 1 <= q and q < r:
            return self.differential_top(r - q).kernel().dimension() - self.differential_top(r - q + 1).rank()
        elif q == 0:
            a = (self.cohomology_group_bottom(0).rank - self.cohomology_group_top(r + 1).rank
                 + self.cohomology_group_top(r).rank)
            b = self.differential_bottom(1).rank() + self.differential_top(r).rank()
            return a - b
        elif q > r:
            return Integer(0)
        raise IndexError('index must be non-negative')
